from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from loguru import logger
from pydantic import BaseModel


class Feedback(BaseModel):
    """Dataclass for feedback given by a user from a UI component."""

    title: str
    description: str
    tags: Optional[List[str]]
    metadata: Optional[Dict[str, Union[List[Any], float, int, str, dict]]]

    def save_local(self, path: str, file_name: Optional[str] = None):
        if path is None:
            raise TypeError("Specify the local path where you would like to save your Trubric json.")
        if file_name is None:
            file_name = "feedback.json"
        with open(Path(path) / file_name, "w") as file:
            file.write(self.json(indent=4))
            logger.info(f"Feedback saved to {Path(path) / file_name}.")



import importlib.util
import json
import os
import sys
from pathlib import Path
from typing import Optional

import typer
from rich import print as rprint

from trubrics.exceptions import MissingConfigPathError, MissingTrubricRunFileError
from trubrics.utils.trubrics_manager_connector import make_request
from trubrics.validations.run import TrubricRun, run_trubric

app = typer.Typer()


@app.command()
def run(
    save_ui: bool = False,
    trubric_config_path: str = typer.Option(
        ...,
        prompt=(
            "Enter the path to your trubric config .json file (this file can be generated by running `trubrics init`)"
        ),
    ),
    trubric_output_file_path: str = typer.Option(
        ".", prompt="Enter a path to save your output trubric file. The default path is"
    ),
    trubric_output_file_name: str = typer.Option(
        "my_new_trubric.json", prompt="Enter the file name of your output trubric file. The default file name is"
    ),
):
    """The CLI `trubrics run` command for running trubrics.

    Args:
        trubric_config_path: path to your trubric config .json file \
            (this file can be generated by running `trubrics init`).
        trubric_output_file_path: path to save your output trubric file
        trubric_output_file_name: file name of your output trubric file.
    """
    trubric_config_file_path = Path(trubric_config_path) / ".trubrics_config.json"
    if not trubric_config_file_path.exists():
        raise MissingConfigPathError(
            f"The trubric configuration file '{trubric_config_file_path}' has not been found. Run `trubrics init` to"
            " create file."
        )
    with open(trubric_config_file_path, "r") as file:
        trubrics_config = json.load(file)
    trubric_run_path = trubrics_config["trubric_run_path"]

    tc = _import_module(module_path=trubric_run_path)
    if hasattr(tc, "RUN_CONTEXT"):
        if isinstance(tc.RUN_CONTEXT, TrubricRun):
            run_context = tc.RUN_CONTEXT
        else:
            raise TypeError("'RUN_CONTEXT' attribute must be of type TrubricRun.")
    else:
        raise AttributeError("Trubrics config python module must contain an attribute 'RUN_CONTEXT'.")

    typer.echo(
        typer.style(
            f"Running trubric from file '{trubric_run_path}' with model '{run_context.trubric.model_name}' and"
            f" dataset '{tc.data_context.name}'.",
            fg=typer.colors.BLUE,
        )
    )
    all_validation_results = run_trubric(tr=run_context)
    validations = []
    for validation_result in all_validation_results:
        validations.append(validation_result)

        message_start = f"{validation_result.validation_type} - {validation_result.severity.upper()}"
        completed_dots = (100 - len(message_start)) * "."
        if validation_result.outcome == "pass":
            ending = typer.style("PASSED", fg=typer.colors.GREEN, bold=True)
        else:
            ending = typer.style("FAILED", fg=typer.colors.WHITE, bg=typer.colors.RED)
        message = typer.style(message_start, bold=True) + completed_dots + ending
        typer.echo(message)

    # save new trubric .json
    new_trubric = tc.trubric
    new_trubric.validations = validations
    new_trubric.save_local(path=trubric_output_file_path, file_name=trubric_output_file_name)

    # save new trubric to ui
    if save_ui is True:
        if "user_id" in trubrics_config.keys():
            new_trubric.save_ui(url=trubrics_config["api_url"], user_id=trubrics_config["user_id"])
        else:
            typer.echo(
                typer.style(
                    "ERROR: You must authenticate with the trubrics manager by running `trubrics init` to remotely save"
                    " trubrics runs.",
                    fg=typer.colors.RED,
                )
            )


@app.command()
def init(
    trubrics_api_url: Optional[str] = None,
    trubric_run_path: str = typer.Option(
        ..., prompt="Enter the path to your trubric run .py file (e.g. examples/classification_titanic/trubric_run.py)"
    ),
    trubric_config_path: str = typer.Option(
        ".", prompt="Enter a path to save your .trubrics_config.json. The default path is"
    ),
):
    """The CLI `trubrics init` command for initialising trubrics config.

    Args:
        trubric_run_path: a path towards the .py file that holds code to run a trubric in \
            the `RUN_CONTEXT=TrubricRun(...)`.
        trubric_config_path: a path to save the .trubrics_config.json to
    """

    if trubrics_api_url:
        uid = typer.prompt("Enter your User ID (generated in the trubrics manager)")

        res = make_request(f"{trubrics_api_url}/api/is_user/{uid}", headers={"Content-Type": "application/json"})
        res = json.loads(res)
        if "is_user" in res.keys():
            message = typer.style(res["msg"], fg=typer.colors.RED, bold=True)
            typer.echo(message)
            raise typer.Abort()
        typer.echo(
            typer.style(
                "Trubrics configuration has been set and user is authenticated with the trubrics manager UI:",
                fg=typer.colors.GREEN,
                bold=True,
            )
        )
        res["api_url"] = trubrics_api_url
    else:
        typer.echo(
            typer.style(
                "Trubrics config set without trubrics manager authentication:", fg=typer.colors.GREEN, bold=True
            )
        )
        res = {}

    if not Path(trubric_run_path).exists() or not trubric_run_path.endswith(".py"):
        raise MissingTrubricRunFileError("Trubric run file path does not exist or is not a .py file.")

    res["trubric_run_path"] = trubric_run_path

    rprint(res)
    with open(
        Path(trubric_config_path) / ".trubrics_config.json",
        "w",
    ) as file:
        file.write(json.dumps(res, indent=4))


def _framework_callback(value: str):
    if value not in ["gradio", "streamlit", "dash"]:
        raise typer.BadParameter("Only 'gradio', 'dash' or 'streamlit' frameworks are supported.")
    return value


@app.command()
def example_app(framework: str = typer.Option("streamlit", callback=_framework_callback)):
    """Run the titanic user feedback collector app."""
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, f"../example/app_titanic_{framework}.py")
    if framework == "streamlit":
        import streamlit.cli

        streamlit.cli._main_run(filename)
    elif framework in ["gradio", "dash"]:
        import subprocess

        subprocess.call(["python3", filename])


def _import_module(module_path: str):
    try:
        spec = importlib.util.spec_from_file_location("module.name", module_path)
        lib = importlib.util.module_from_spec(spec)  # type: ignore
        sys.modules["module.name"] = lib
        spec.loader.exec_module(lib)  # type: ignore
    except FileNotFoundError as e:
        raise e
    return lib


if __name__ == "__main__":
    app()



from typing import Any, Dict, List, Optional

import streamlit as st

from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback


def collect_feedback_streamlit(
    path: str,
    file_name: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
):
    """Get user feedback and save."""
    with st.form("form", clear_on_submit=True):
        title = st.text_input(label=config.TITLE, help=config.TITLE_EXPLAIN, key="title")
        description = st.text_input(label=config.DESCRIPTION, help=config.DESCRIPTION_EXPLAIN, key="description")
        submitted = st.form_submit_button(config.FEEDBACK_SAVE_BUTTON)
        if submitted:
            if len(title) == 0 or len(description) == 0:
                st.markdown(
                    config.FEEDBACK_NOT_SAVED_HTML,
                    unsafe_allow_html=True,
                )
            else:
                feedback = Feedback(title=title, description=description, tags=tags, metadata=metadata)
                feedback.save_local(path=path, file_name=file_name)
                st.markdown(
                    config.FEEDBACK_SAVED_HTML,
                    unsafe_allow_html=True,
                )



import gradio as gr

from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback


def collect_feedback_gradio(path: str, file_name=None, tags=None, metadata=None):
    def get_feedback(title: str, description: str):
        if not (len(title) == 0 or len(description) == 0):
            feedback = Feedback(title=title, description=description, tags=tags, metadata=metadata)
            feedback.save_local(path=path, file_name=file_name)
            return config.FEEDBACK_SAVED_HTML
        else:
            return config.FEEDBACK_NOT_SAVED_HTML

    title = gr.Textbox(label=config.TITLE, placeholder=config.TITLE_EXPLAIN)
    description = gr.Textbox(label=config.DESCRIPTION, placeholder=config.DESCRIPTION_EXPLAIN, lines=5)
    feedback_button = gr.Button(config.FEEDBACK_SAVE_BUTTON)
    kwargs = {"fn": get_feedback, "inputs": [title, description], "outputs": gr.HTML()}
    return feedback_button.click(**kwargs)



import gradio as gr

from trubrics.feedback import collect_feedback_gradio

with gr.Blocks() as demo:
    gr.Markdown("Gradio app to collect user feedback on models.")
    with gr.Tab("Feedback"):
        collect_feedback_gradio(path=".")

demo.launch()



TITLE = "Title"
TITLE_EXPLAIN = "Give the issue an explicit title."
DESCRIPTION = "Description"
DESCRIPTION_EXPLAIN = "Detail the issue you have observed."
FEEDBACK_SAVE_BUTTON = "Save feedback"
FEEDBACK_SAVED = "Feedback saved & sent to the Data Science team."
FEEDBACK_SAVED_HTML = f'<p style="color:Green;">{FEEDBACK_SAVED}</p>'
FEEDBACK_NOT_SAVED = "Please specify a feedback title and a description."
FEEDBACK_NOT_SAVED_HTML = '<p style="color:Red;">Please specify a feedback title and a description.</p>'



from trubrics.context import DataContext
from trubrics.validations import ModelValidator
from trubrics.validations.validation_output import (
    validation_output,
    validation_output_type,
)


class CustomValidator(ModelValidator):
    def __init__(self, data: DataContext, model, custom_scorers=None, slicing_functions=None):
        super().__init__(data, model, custom_scorers, slicing_functions)

    def _validate_master_age(self, age_limit_master) -> validation_output_type:
        """
        Write your custom validation function here.

        Notes
        -----
            This method is separated from validate_performance_for_different_fares
            to apply @validation_output and for unit testing.

            The @validation_output decorator allows you to generate a Validation object,
            and must be used to be able to save your validation as part of a Trubric.
            This decorator requires you to return values with the same type as validation_output_type.
        """
        master_df = self.tm.data.testing_data.loc[lambda df: df["Title"] == "Master"]
        errors_df = master_df.loc[lambda df: df["Age"] >= age_limit_master]
        return len(errors_df) == 0, {"errors_df": errors_df.to_dict()}

    @validation_output
    def validate_master_age(self, age_limit_master: int, severity=None):
        """Validate that passengers with the title "master" are younger than a certain age

        Args:
            age_limit_master: cut off value for master

        Returns:
            True for success, false otherwise. With a results dictionary giving dict of errors.
        """
        return self._validate_master_age(age_limit_master)



import pandas as pd
import streamlit as st

from trubrics.context import DataContext


def generate_what_if_streamlit(data_context: DataContext):
    """
    Generate a what-if tool based on the testing_data of the DataContext.
    """
    out_df = pd.DataFrame()
    wi_data = data_context.testing_data
    for col in wi_data.columns:
        series = wi_data[col]
        if data_context.business_columns is not None and col in data_context.business_columns.keys():
            renamed_col = data_context.business_columns[col]
        else:
            renamed_col = col
        if data_context.categorical_columns is None:
            raise ValueError(
                "A list of categorical_columns must be specified in the DataContext for the what-if functionality."
            )
        if col in data_context.categorical_columns:
            out_df[col] = [st.selectbox(renamed_col, tuple(series.dropna().unique()))]
        else:
            out_df[col] = [
                st.slider(
                    renamed_col,
                    min_value=int(series.min()),
                    max_value=int(series.max()),
                    value=round(series.mean()),
                )
            ]
    return out_df


def explore_testing_data(data_context, model):
    data_view = st.selectbox(label="", options=("View full test set", "View test set errors", "View split by target"))
    if data_view == "View full test set":
        st.dataframe(data_context.renamed_testing_data)
    elif data_view == "View test set errors":
        st.dataframe(
            _filter_testing_data_errors(data_context=data_context, model=model).rename(data_context.business_columns)
        )
    elif data_view == "View split by target":
        target_split = st.radio(label="", options=data_context.testing_data[data_context.target].unique())
        st.dataframe(data_context.testing_data.loc[lambda x: x[data_context.target] == target_split])


def _filter_testing_data_errors(data_context, model):
    predict_col = f"{data_context.target}_predictions"
    assign_kwargs = {predict_col: model.predict(data_context.testing_data)}
    return data_context.testing_data.assign(**assign_kwargs).loc[lambda x: x[data_context.target] != x[predict_col], :]



from trubrics.feedback.collect.dash import collect_feedback_dash
from trubrics.feedback.collect.gradio import collect_feedback_gradio
from trubrics.feedback.collect.streamlit import collect_feedback_streamlit
from trubrics.feedback.experiment.streamlit import (
    explore_testing_data,
    generate_what_if_streamlit,
)

__all__ = [
    "collect_feedback_dash",
    "collect_feedback_gradio",
    "collect_feedback_streamlit",
    "explore_testing_data",
    "generate_what_if_streamlit",
]



from typing import Any, Dict, List, Optional

import dash_bootstrap_components as dbc
from dash import Input, Output, callback, callback_context, html

from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback


def collect_feedback_dash(
    path: str,
    file_name: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
):
    title_input = html.Div(
        [
            dbc.Label(config.TITLE, html_for="title"),
            dbc.Input(id="title", placeholder=config.TITLE_EXPLAIN),
        ],
        className="mb-3",
    )

    description_input = html.Div(
        [
            dbc.Label(config.DESCRIPTION, html_for="description"),
            dbc.Input(id="description", placeholder=config.DESCRIPTION_EXPLAIN),
        ],
        className="mb-3",
    )

    button = html.Div(
        [
            html.Div(),
            dbc.Button(config.FEEDBACK_SAVE_BUTTON, id="button_input", color="secondary"),
            html.P(id="message"),
        ],
        className="mb-3",
    )

    @callback(
        Output("message", "children"),
        Output("message", "style"),
        Output("title", "value"),
        Output("description", "value"),
        Input("button_input", "n_clicks"),
        Input("title", "value"),
        Input("description", "value"),
    )
    def on_button_click(n, title, description):
        changed_id = [p["prop_id"] for p in callback_context.triggered][0]
        if "button_input" in changed_id:
            if title is None or description is None:
                return config.FEEDBACK_NOT_SAVED, {"color": "Red"}, title, description
            else:
                feedback = Feedback(title=title, description=description, tags=tags, metadata=metadata)
                feedback.save_local(path=path, file_name=file_name)
                return config.FEEDBACK_SAVED, {"color": "Green"}, None, None
        else:
            return None, None, title, description

    return dbc.Form([title_input, description_input, button])



import streamlit as st

from trubrics.context import DataContext
from trubrics.example import get_titanic_data_and_model
from trubrics.example import titanic_config as tc
from trubrics.feedback import (
    collect_feedback_streamlit,
    explore_testing_data,
    generate_what_if_streamlit,
)

_, test_df, model = get_titanic_data_and_model()

data_context = DataContext(
    testing_data=test_df,
    target="Survived",
    categorical_columns=tc.CATEGORICAL_COLUMNS,
    business_columns=tc.BUSINESS_COLUMNS,
)

with st.sidebar:
    st.title("Test the model with different inputs")
    df = generate_what_if_streamlit(data_context=data_context)
wi_prediction = model.predict(df)[0]

st.title("View model prediction")
if wi_prediction:
    prediction = '<p style="color:Green;">This passenger would have survived.</p>'
else:
    prediction = '<p style="color:Red;">This passenger would have died.</p>'
st.markdown(prediction, unsafe_allow_html=True)

st.title("Send model feedback")
metadata = {"what_if_data": df.to_dict(), "what_if_prediction": wi_prediction}
collect_feedback_streamlit(path=".", metadata=metadata)

st.title("View data")
explore_testing_data(data_context=data_context, model=model)
